;DCF77 Date Display with 14-segment LED display for weekday
;Controller: AT90S2313 or ATTiny2313 @ 4.194304 MHz external crystal
;Fuses for ATTiny2313: low=0xDD, high=0xDB, ext=0xFF
;
;Assemble with avra (http://avra.sourceforge.net/) only, this file contains
;directives specific for this assembler!
;
;(C) Arne Rossius, http://www.ebps.de.vu/
;
;DISPLAY:
;  - LED displays, 14-segment dual-digit + 4x 7-segment, common anode
;  - 16-bit shift register for segments a..m,dp,NC / 1s[a..g,dp], 10s[a..g,dp]
;  - shift register outputs are active low (common anode displays)
;  - display order: WW MM-DD
;
;PORTS:
;  PB0 = Shift Register Data
;  PB1 = Shift Register Clock
;  PB2 = Shift Register Strobe
;  PB4 = Display 1 (Weekday first letter) Anode (active low)
;  PB5 = Display 2 (Weekday second letter) Anode (active low)
;  PB6 = Display 3/4 (Month) Anodes (active low)
;  PB7 = Display 5/6 (Day) Anodes (active low)
;  PD2 = DCF77 input
;  PD6 = "disable display" switch (open: off, closed to GND: on)
;
;VERSION HISTORY:
;  2008-06-08  first version


;microcontroller definition file
.include "../include/2313def.inc"

;DCF77 receiver
.equ	DCFPIN = PIND
.equ	DCFInput = 2
.equ	DCFInvert = 1 ;inverted output (negative pulses)

;"Disable Display" switch
.equ	SwitchPIN = PIND
.equ	DisplayOff = 6

;LED display
.equ	DisplayPORT = PORTB
.equ	DisplayData = 0
.equ	DisplayClock = 1
.equ	DisplayStrobe = 2



;needed for lpm instruction
.def	lpmresult = R0

;registers for DCF77 telegram
.def	dcf1 = R3
.def	dcf2 = R4
.def	dcf3 = R5
.def	dcf4 = R6
.def	dcf5 = R7

;time/date registers
.def	subsec = R8
.def	second = R9
.def	minute = R10
.def	hour = R11
.def	day = R12
.def	month = R13
.def	year = R14
.def	wkday = R15 ;(weekday, 1=MO 7=SU)

;general registers
.def	temp = R16
.def	temp2 = R17

;DCF77-specific registers
.def	dcfbits = R18 ;received bits
.def	dcftime = R19 ;1/64-seconds since last edge
.def	flags1 = R20
.def	flags2 = R21

;additional registers for display
.def	display = R22 ;active display output mask
.def	byte1 = R23 ;first byte for active display (a..g2 / 1s a..g, dp)
.def	byte2 = R24 ;second byte for active display (h..m, dp / 10s a..g, dp)

;bit definitions for the flags1 register
.equ	ErrBitCount = 0 ;last error: wrong number of bits
.equ	ErrParity = 1 ;last error: parity error
.equ	ErrBitLength = 2 ;last error: wrong bit length
.equ	ErrStartBit = 3 ;last error: start bit was not 1
.equ	TimeOK = 4 ;time is OK
.equ	DateOK = 5 ;date is OK
.equ	StopDCF = 6 ;stop DCF77 reception until a new telegram starts
.equ	DCFdisabled = 7 ;DCF77 receiver disabled (reset when a new hour begins)
.equ	ErrMask = 1<<ErrBitLength | 1<<ErrParity | 1<<ErrBitCount | 1<<ErrStartBit ;used to reset errors

;bit definitions for the flags2 register
.equ	LastState = DCFInput ;last state of DCF77 input, must be the same bit as on the port!

.cseg

;*******************************************************************************
;*
;*   MAIN PROGRAM
;*
;*******************************************************************************

;---------------------------------------
; Interrupt vectors
;---------------------------------------

;Reset
.org 0x000
	rjmp	start
	
;Timer 0 overflow
.org OVF0addr
	rjmp	ovf0int

;---------------------------------------
; Length of every month in days
;---------------------------------------

months:
	;length of months + 1, packed BCD
	;    Jan   Feb   Mar   Apr   May   Jun   Jul   Aug   Sep   Oct   Nov   Dec
	.db 0x32, 0x29, 0x32, 0x31, 0x32, 0x31, 0x32, 0x32, 0x31, 0x32, 0x31, 0x32 ;normal
	.db 0x32, 0x30, 0x32, 0x31, 0x32, 0x31, 0x32, 0x32, 0x31, 0x32, 0x31, 0x32 ;leap year

;---------------------------------------
; Initialization
;---------------------------------------
	
start:
	;stack pointer
	ldi	temp, LOW(RAMEND)
	out	SPL, temp
	
	;disable analog comparator
	ldi	temp, 0x80
	out	ACSR, temp
	
	;I/O ports
	ldi	temp, 0xFF
	out	DDRB, temp
	ldi	temp, 0xF0
	out	PORTB, temp
	ldi	temp, 0x00
	out	DDRD, temp
	ldi	temp, 0xFF
	out	PORTD, temp
	
	;timer
	ldi	temp, 1<<TOIE0 ;overflow int on
	out	TIMSK, temp
	ldi	temp, 0x04 ;Clk/256 = 64Hz int
	out	TCCR0, temp
	
	;registers
	ldi	dcfbits, 0
	ldi	dcftime, 0
	ldi	flags1, 1<<StopDCF
	ldi	flags2, 0
	ldi	display, 0xE0
	
	;enable interrupts
	sei
	
;---------------------------------------
; Main loop
;   modifies: temp, temp2, display, ZL, ZH, lpmresult
;---------------------------------------

;multiplex all the time and do *all* the action while one of the displays is enabled
main_loop:
	;enable next display
	ldi	temp, 0xF0
	out	DisplayPORT, temp
	sbi	DisplayPORT, DisplayStrobe
	cbi	DisplayPORT, DisplayStrobe
	sbis	SwitchPIN, DisplayOff
	out	DisplayPORT, display
	
	;encode the next display
	lsl	display
	ori	display, 0x10
	cpi	display, 0xF0
	brne	PC+2
	ldi	display, 0xE0
	
	sbrs	display, 6 ;3rd/4th digits: month
	rjmp	disp_month
	sbrs	display, 7 ;5th/6th digits: day
	rjmp	disp_day
	
	;1st/2nd digits: weekday
	ldi	ZL, LOW(weekdays_14seg*2)
	ldi	ZH, HIGH(weekdays_14seg*2)
	sbrs	display, 5
	adiw	ZH:ZL, 14
	mov	temp, wkday
	dec	temp
	lsl	temp
	add	ZL, temp
	ldi	temp, 0
	adc	ZH, temp
	lpm
	mov	byte1, lpmresult
	adiw	ZH:ZL, 1
	lpm
	mov	byte2, lpmresult
	sbrc	flags1, TimeOK
	rjmp	shift
	sbrs	display, 5
	rjmp	disp_lastDCFbit
	ldi	byte1, 0x08 ;'_'
	ldi	byte2, 0x00
	sbrc	flags2, LastState
	ldi	byte1, 0xC0 ;'-'
	rjmp	shift
disp_lastDCFbit:
	ldi	byte1, 0x3F ;'0'
	ldi	byte2, 0x0C
	sbrc	dcf5, 7
	rjmp	shift
	ldi	byte1, 0x06 ;'1'
	ldi	byte2, 0x04
	rjmp	shift
	
disp_day:
	mov	temp, day
	mov	byte2, lpmresult
	sbrc	flags1, TimeOK
	rjmp	encode_7seg
	ldi	temp, 0xFE ;" -"
	sbrc	flags1, ErrBitCount
	ldi	temp, 0xFA ;" C"
	sbrc	flags1, ErrBitLength
	ldi	temp, 0xFB ;" L"
	sbrc	flags1, ErrParity
	ldi	temp, 0xFC ;" P"
	sbrc	flags1, ErrStartBit
	ldi	temp, 0xFD ;" S"
	rjmp	encode_7seg

disp_month:
	mov	temp, month
	sbrc	flags1, TimeOK
	rjmp	encode_7seg
	mov	temp, dcfbits
	rcall	bin2bcd
	
encode_7seg:
	ldi	ZL, LOW(numbers_7seg*2)
	ldi	ZH, HIGH(numbers_7seg*2)
	mov	temp2, temp
	swap	temp2
	andi	temp2, 0x0F
	add	ZL, temp2
	ldi	temp2, 0
	add	ZH, temp2
	lpm
	mov	byte1, lpmresult
	ldi	ZL, LOW(numbers_7seg*2)
	ldi	ZH, HIGH(numbers_7seg*2)
	andi	temp, 0x0F
	add	ZL, temp
	ldi	temp, 0
	add	ZH, temp
	lpm
	mov	byte2, lpmresult
	
shift:
	;shift out 16 bits, inverted
	ldi	temp, 16
shift_loop:
	sbi	DisplayPORT, DisplayData
	lsl	byte1
	rol	byte2
	brcc	PC+2
	cbi	DisplayPORT, DisplayData
	sbi	DisplayPORT, DisplayClock
	cbi	DisplayPORT, DisplayClock
	dec	temp
	brne	shift_loop
	
	;short delay
	ldi	ZH, 2
	ldi	ZL, 0
multiplex_wait:
	dec	ZL
	brne	multiplex_wait
	dec	ZH
	brne	multiplex_wait
	
	rjmp	main_loop
	
;---------------------------------------
; 7- and 14-segment encoding tables
;---------------------------------------

numbers_7seg:
	;     0     1     2     3     4     5     6     7
	.db 0x3F, 0x06, 0x5B, 0x4F, 0x66, 0x6D, 0x7D, 0x07
	;     8     9     C     L     P     S     -    SPC
	.db 0x7F, 0x6F, 0x39, 0x38, 0x73, 0x6D, 0x40, 0x00
	
weekdays_14seg:
	;German
	;      M       D       M       D       F       S       S
	.dw 0x0536, 0x120F, 0x0536, 0x120F, 0x00F1, 0x00ED, 0x00ED
	;      O       I       I       O       R       A       O
	.dw 0x003F, 0x1209, 0x1209, 0x003F, 0x20F3, 0x00F7, 0x003F
	
;	;English
;	;      M       T       W       T       F       S       S
;	.dw 0x0536, 0x1201, 0x2436, 0x1201, 0x00F1, 0x00ED, 0x00ED
;	;      O       U       E       H       R       A       U
;	.dw 0x003F, 0x003E, 0x00F9, 0x00F6, 0x20F3, 0x00F7, 0x003E

;*******************************************************************************
;*
;*   DCF77 subroutines
;*
;*******************************************************************************

;---------------------------------------
; Timer 0 overflow int (64Hz)
;   modifies: dcftime, dcfbits, flags1, flags2, dcf1, dcf2, dcf3, dcf4, dcf5,
;             subsec, second, minute, hour, day, wkday, month, year
;---------------------------------------

ovf0int:
	;save registers
	push	temp
	push	temp2
	in	temp, SREG
	push	temp
	
	;don't evaluate DCF77 signal when reception is disabled
	sbrc	flags1, DCFdisabled
	rjmp	dcf_end
	
	inc	dcftime
	
	;compare DCF77 input to last saved state
	in	temp, DCFPIN
	eor	temp, flags2
	andi	temp, 1<<LastState
	.if ( DCFInvert )
		.message "DCF77 Signal: INVERTED"
		brne	dcf_end ;no change
	.else
		.message "DCF77 Signal: normal"
		breq	dcf_end ;no change
	.endif
	
	;take appropriate action for pos/neg edges
	sbrs	flags2, LastState
	rjmp	dcf_posedge
	
		; < negative edge ( 1 -> 0 ) >
	
	andi	flags2, ~(1<<LastState)
	
	;exit if an error was detected before
	sbrc	flags1, StopDCF
	rjmp	dcf_end
	 
	;increment bit number, exit if more than 59 bits were already received
	inc	dcfbits
	cpi	dcfbits, 60
	brsh	PC+2
	rjmp	dcf_evalbit
	andi	flags1, ~ErrMask
	ori	flags1, 1<<ErrBitCount | 1<<StopDCF
	rjmp	dcf_end
	
		; < positive edge ( 0 -> 1 ) >
dcf_posedge:
	ori	flags2, 1<<LastState
	
	;check if time was >1.2s (start of new telegram)
	cpi	dcftime, 77
	ldi	dcftime, 0
	brlo	dcf_end
	
	;end of telegram / new telegram
	andi	flags1, ~ErrMask & ~(1<<StopDCF)
	cpi	dcfbits, 59
	ldi	dcfbits, 0
	brne	PC+2
	rjmp	dcf_eval
	ori	flags1, 1<<ErrBitCount
dcf_end:

;---------------------------------------
; increment time (fallback quartz clock)
;---------------------------------------

	;increment 1/64-seconds
	inc	subsec
	ldi	temp, 64
	cp	subsec, temp
	brlo	inctime_end
	
	;increment seconds
	clr	subsec
	mov	temp, second
	rcall	bcdinc
	mov	second, temp
	cpi	temp, 0x60
	brlo	inctime_end
	
	;increment minutes
	clr	second
	mov	temp, minute
	rcall	bcdinc
	mov	minute, temp
	cpi	temp, 0x04 ;disable DCF77 reception if nothing received for 4 minutes
	brne	inctime_nodisable
	sbrc	flags1, TimeOK
	ori	flags1, 1<<DCFdisabled
inctime_nodisable:
	cpi	temp, 0x60
	brlo	inctime_end

	;increment hours
	clr	minute
	andi	flags1, ~(1<<DCFdisabled) ;re-enable DCF77 reception
	mov	temp, hour
	rcall	bcdinc
	mov	hour, temp
	cpi	temp, 0x24
	brlo	inctime_end
	
	;increment weekday and days
	clr	hour
	mov	temp, wkday
	inc	temp
	cpi	temp, 8
	brlo	PC+2
	ldi	temp, 1
	mov	wkday, temp
	ldi	ZL, LOW(months*2-1) ;get length of current month
	ldi	ZH, HIGH(months*2-1)
	sbrc	year, 0
	rjmp	noleap
	sbrc	year, 1
	rjmp	noleap
	adiw	ZH:ZL, 12
noleap:
	mov	temp, month
	rcall	bcd2bin
	add	ZL, temp
	ldi	temp, 0
	adc	ZH, temp
	lpm
	mov	temp, day ;increment day
	rcall	bcdinc
	mov	day, temp
	cp	temp, R0 ;R0 is result from LPM
	brlo	inctime_end
	
	;increment month
	clr	day
	inc	day
	mov	temp, month
	rcall	bcdinc
	mov	month, temp
	cpi	temp, 0x13
	brlo	inctime_end
	
	;increment year
	clr	month
	inc	month
	mov	temp, year
	rcall	bcdinc
	mov	year, temp
inctime_end:

	pop	temp
	out	SREG, temp
	pop	temp2
	pop	temp
	reti

;---------------------------------------
; DCF77 reception (continued)
;---------------------------------------

		; < negative edge - cont'd >
	
	;check bit length (optional)
dcf_evalbit:
	cpi	dcftime, 5 ;bit is 0: 5 <= dcftime <= 7
	brlo	dcf_errbit
	cpi	dcftime, 8
	brlo	dcf_clearbit
	cpi	dcftime, 11 ;bit is 1: 11 <= dcftime <= 13
	brlo	dcf_errbit
	cpi	dcftime, 14
	brlo	dcf_setbit
	
	;received bit is erroneous
dcf_errbit:
	andi	flags1, ~ErrMask
	ori	flags1, 1<<ErrBitLength | 1<<StopDCF
	rjmp	dcf_end
	
	;save bit
dcf_clearbit:
	clc ;clear bit
	rjmp	dcf_savebit
dcf_setbit:
	sec ;set bit
dcf_savebit:
	ror	dcf5
	ror	dcf4
	ror	dcf3
	ror	dcf2
	ror	dcf1
	rjmp	dcf_end
	
;---------------------------------------
; evaluate DCF77 telegram
;---------------------------------------
	
dcf_eval:
	;check date parity
	ldi	temp2, 0
	mov	temp, dcf3
	andi	temp, 0xFE
	rcall	parity
	mov	temp, dcf4
	rcall	parity
	mov	temp, dcf5
	andi	temp, 0x7F
	rcall	parity
	mov	temp, dcf5
	lsl	temp ;Bit7 -> C
	rol	temp ;C -> Bit0
	andi	temp, 0x01
	cp	temp, temp2
	brne	dcf_parityerror
	
	;check hour parity
	ldi	temp2, 0
	mov	temp, dcf2
	andi	temp, 0xFC
	rcall	parity
	mov	temp, dcf3
	andi	temp, 0x01
	cp	temp, temp2
	brne	dcf_parityerror
	
	;check minute parity
	ldi	temp2, 0
	mov	temp, dcf1
	andi	temp, 0xFC
	rcall	parity
	mov	temp, dcf2
	andi	temp, 0x01
	rcall	parity
	mov	temp, dcf2
	lsr	temp
	andi	temp, 0x01
	cp	temp, temp2
	brne	dcf_parityerror

	;save minutes
	lsr	dcf2
	ror	dcf1
	lsr	dcf1
	mov	minute, dcf1
	
	;save hours
	lsr	dcf2
	mov	hour, dcf2
	
	;save day
	lsl	dcf3 ;align day
	rol	dcf4 ;align month[5] weekday[3]
	rol	dcf5 ;align year
	lsr	dcf3
	lsr	dcf3
	mov	day, dcf3
	
	;save weekday
	mov	wkday, dcf4
	ldi	temp, 0x07
	and	wkday, temp
	
	;save month
	lsr	dcf4
	lsr	dcf4
	lsr	dcf4
	mov	month, dcf4
	
	;save year
	mov	year, dcf5
	
	;set time & date ok, disable DCF77 reception
	clr	subsec
	clr	second
	ori	flags1, 1<<TimeOK | 1<<DateOK | 1<<DCFdisabled
	rjmp	dcf_end
	
dcf_parityerror:
	;set parity error
	ori	flags1, 1<<ErrParity
	rjmp	dcf_end

;*******************************************************************************
;*
;*   GENERAL ARITHMETIC FUNCTIONS
;*
;*******************************************************************************

;---------------------------------------
; calculate parity and add up
;   in: temp
;   out: temp2 (add to existing value)
;---------------------------------------

parity:
	lsr	temp
	brcc	PC+2
	inc	temp2
	cpi	temp, 0
	brne	parity
	andi	temp2, 0x01
	ret
	
;---------------------------------------
; increment packed bcd number
;   in: temp
;   out: temp
;---------------------------------------

bcdinc:
	subi	temp, -7 ;increment and add 6
	brhc	PC+2 ;half-carry is inverted when using "subi rh, -k" instead of add!
	subi	temp, 6
	ret
	
;---------------------------------------
; convert binary to packed BCD
;   in: temp
;   out: temp
;   modifies: temp2
;---------------------------------------

bin2bcd:
	ldi	temp2, 0
bin2bcd_loop:
	subi	temp2, -0x10
	subi	temp, 10
	brcc	bin2bcd_loop
	subi	temp2, 0x10
	subi	temp, -10
	add	temp, temp2
	ret
	
;---------------------------------------
; convert packed BCD to binary
;   in: temp
;   out: temp
;   modifies: temp2
;---------------------------------------

bcd2bin:
	mov	temp2, temp
	andi	temp, 0x0F
	swap	temp2
	andi	temp2, 0x0F
	lsl	temp2
	add	temp, temp2 ;temp += temp2*2
	lsl	temp2
	lsl	temp2
	add	temp, temp2 ;temp += temp2*8
	ret
